/* ====================================================================
   Licensed to the Apache Software Foundation (ASF) under one or more
   contributor license agreements.  See the NOTICE file distributed with
   this work for Additional information regarding copyright ownership.
   The ASF licenses this file to You under the Apache License, Version 2.0
   (the "License"); you may not use this file except in compliance with
   the License.  You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
==================================================================== */

namespace TestCases.HSSF.Model
{
    using System;
    using NUnit.Framework;
    using Npoi.Core.HSSF.Record;
    using Npoi.Core.HSSF.UserModel;
    using Npoi.Core.SS.Formula;
    using Npoi.Core.SS.Formula.Eval;
    using Npoi.Core.SS.Formula.Functions;
    using Npoi.Core.SS.Formula.Udf;
    using TestCases.HSSF.UserModel;
    using Npoi.Core.HSSF.Model;

    /**
     * Unit test for the Workbook class.
     *
     * @author Glen Stampoultzis (glens at apache.org)
     */
    [TestFixture]
    public class TestWorkbook
    {
        [Test]
        public void TestFontStuff()
        {
            InternalWorkbook wb = TestHSSFWorkbook.GetInternalWorkbook(new HSSFWorkbook());

            Assert.AreEqual(4, wb.NumberOfFontRecords);
            Assert.AreEqual(68, wb.Records.Count);

            FontRecord f1 = wb.GetFontRecordAt(0);
            FontRecord f4 = wb.GetFontRecordAt(3);

            Assert.AreEqual(0, wb.GetFontIndex(f1));
            Assert.AreEqual(3, wb.GetFontIndex(f4));

            Assert.AreEqual(f1, wb.GetFontRecordAt(0));
            Assert.AreEqual(f4, wb.GetFontRecordAt(3));

            // There is no 4! new ones go in at 5

            FontRecord n = wb.CreateNewFont();
            Assert.AreEqual(69, wb.Records.Count);
            Assert.AreEqual(5, wb.NumberOfFontRecords);
            Assert.AreEqual(5, wb.GetFontIndex(n));
            Assert.AreEqual(n, wb.GetFontRecordAt(5));

            // And another
            FontRecord n6 = wb.CreateNewFont();
            Assert.AreEqual(70, wb.Records.Count);
            Assert.AreEqual(6, wb.NumberOfFontRecords);
            Assert.AreEqual(6, wb.GetFontIndex(n6));
            Assert.AreEqual(n6, wb.GetFontRecordAt(6));


            // Now remove the one formerly at 5
            Assert.AreEqual(70, wb.Records.Count);
            wb.RemoveFontRecord(n);

            // Check that 6 has gone to 5
            Assert.AreEqual(69, wb.Records.Count);
            Assert.AreEqual(5, wb.NumberOfFontRecords);
            Assert.AreEqual(5, wb.GetFontIndex(n6));
            Assert.AreEqual(n6, wb.GetFontRecordAt(5));

            // Check that the earlier ones are unChanged
            Assert.AreEqual(0, wb.GetFontIndex(f1));
            Assert.AreEqual(3, wb.GetFontIndex(f4));
            Assert.AreEqual(f1, wb.GetFontRecordAt(0));
            Assert.AreEqual(f4, wb.GetFontRecordAt(3));

            // Finally, add another one
            FontRecord n7 = wb.CreateNewFont();
            Assert.AreEqual(70, wb.Records.Count);
            Assert.AreEqual(6, wb.NumberOfFontRecords);
            Assert.AreEqual(6, wb.GetFontIndex(n7));
            Assert.AreEqual(n7, wb.GetFontRecordAt(6));
        }
        private class FreeRefFunction1 : FreeRefFunction
        {
            #region FreeRefFunction Ա

            public ValueEval Evaluate(ValueEval[] args, OperationEvaluationContext ec)
            {
                throw new NotImplementedException("not implemented");
            }

            #endregion
        }
        [Test]
        public void TestAddNameX()
        {
            InternalWorkbook wb = TestHSSFWorkbook.GetInternalWorkbook(new HSSFWorkbook());
            Assert.IsNotNull(wb.GetNameXPtg("ISODD", UDFFinder.DEFAULT));

            FreeRefFunction1 NotImplemented = new FreeRefFunction1();

            /**
             * register the two test UDFs in a UDF Finder, to be passed to the Evaluator
             */
            UDFFinder udff1 = new DefaultUDFFinder(new string[] { "myFunc", },
                    new FreeRefFunction[] { NotImplemented });
            UDFFinder udff2 = new DefaultUDFFinder(new string[] { "myFunc2", },
                    new FreeRefFunction[] { NotImplemented });
            UDFFinder udff = new AggregatingUDFFinder(udff1, udff2);
            Assert.IsNotNull(wb.GetNameXPtg("myFunc", udff));
            Assert.IsNotNull(wb.GetNameXPtg("myFunc2", udff));

            Assert.IsNull(wb.GetNameXPtg("myFunc3", udff));  // myFunc3 is unknown
        }
        [Test]
        public void TestRecalcId()
        {
            HSSFWorkbook wb = new HSSFWorkbook();
            Assert.IsFalse(wb.ForceFormulaRecalculation);

            InternalWorkbook iwb = TestHSSFWorkbook.GetInternalWorkbook(wb);
            int countryPos = iwb.FindFirstRecordLocBySid(CountryRecord.sid);
            Assert.IsTrue(countryPos != -1);
            // RecalcIdRecord is not present in new workbooks
            Assert.AreEqual(null, iwb.FindFirstRecordBySid(RecalcIdRecord.sid));
            RecalcIdRecord record = iwb.RecalcId;
            Assert.IsNotNull(record);
            Assert.AreSame(record, iwb.RecalcId);

            Assert.AreSame(record, iwb.FindFirstRecordBySid(RecalcIdRecord.sid));
            Assert.AreEqual(countryPos + 1, iwb.FindFirstRecordLocBySid(RecalcIdRecord.sid));

            record.EngineId = (/*setter*/100);
            Assert.AreEqual(100, record.EngineId);
            Assert.IsTrue(wb.ForceFormulaRecalculation);

            wb.ForceFormulaRecalculation = (/*setter*/true); // resets the EngineId flag to zero
            Assert.AreEqual(0, record.EngineId);
            Assert.IsFalse(wb.ForceFormulaRecalculation);
        }
    }

}